/*            Copyright (C) 2000, 2001, 2002, 2003 Stijn van Dongen
 *
 * This file is part of MCL.  You can redistribute and/or modify MCL under the
 * terms of the GNU General Public License; either version 2 of the License or
 * (at your option) any later version.  You should have received a copy of the
 * GPL along with MCL, in the file COPYING.
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <ctype.h>


#include "compile.h"
#include "io.h"
#include "types.h"
#include "err.h"
#include "alloc.h"


static void mcxIOreset
(  mcxIO*    xf
)  ;


static int mcxIOwarnfp
(  mcxIO*    xf
,  const char* who
)
   {  if (xf->fp && !xf->stdio)
      {  mcxIOerr(xf, who, "has open file pointer")
      ;  return 1
   ;  }
      return 0
;  }


void mcxIOclose
(  mcxIO*    xf
)
   {  if (xf->fp && !xf->stdio)
      {  fclose(xf->fp)
      ;  xf->fp = NULL
      ;  mcxIOreset(xf)
   ;  }
   }


/* note: does not touch all members, notably
 *    ufo
 *    fn
 *    mode
 *    fp
*/

static void mcxIOreset
(  mcxIO*    xf
)
   {  xf->lc      =  1
   ;  xf->lo      =  0
   ;  xf->bc      =  0
   ;  xf->ateof   =  0
   ;  xf->stdio   =  0
;  }


void mcxIOerr
(  mcxIO*   xf
,  const char     *complainer
,  const char     *complaint
)
   {  if (!xf)
      return

   ;  mcxErr
      (  complainer
      ,  "%s stream <%s> %s"
      ,  xf->mode
      ,  xf->fn->str
      ,  complaint
      )  
;  }


mcxIO* mcxIOnew
(  const char*       str
,  const char*       mode
)
   {  if (!str || !mode)
      {  mcxErr("mcxIOnew PBD", "void string or mode argument")
      ;  return NULL
   ;  }

      return mcxIOrenew(NULL, str, mode)
;  }


mcxIO* mcxIOrenew
(  mcxIO*            xf
,  const char*       name
,  const char*       mode
)
   {  if
      (  mode &&
         (  !strstr(mode, "w") && !strstr(mode, "r") && !strstr(mode, "a") )
      )
      {  mcxErr ("mcxIOrenew PBD", "unsupported open mode <%s>", mode)
      ;  return NULL
   ;  }

      if (!xf)
      {  if (!name || !mode)
         {  mcxErr ("mcxIOrenew PBD", "too few arguments")
         ;  return NULL
      ;  }
         xf = (mcxIO*) mcxAlloc(sizeof(mcxIO), EXIT_ON_FAIL)
      ;  xf->fn   =  mcxTingEmpty(NULL, 20)
      ;  xf->ufo  =  NULL
      ;  xf->mode =  "_"
      ;  xf->fp   =  NULL
   ;  }
      else if (mcxIOwarnfp(xf, "mcxIOrenew"))
      mcxIOclose(xf)

   ;  mcxIOreset(xf)

   ;  if (name)
      mcxTingWrite(xf->fn, name)

   ;  if (mode)
      xf->mode = mode

   ;  return xf
;  }


mcxstatus  mcxIOopen
(  mcxIO*   xf
,  mcxOnFail      ON_FAIL
)
   {  const char* treat    =  ""
   ;  const char* fname    =  xf->fn->str
   ;  if (!xf)
      {  mcxErr("mcxIOnew PBD", "received void object")
      ;  mcxFail()
   ;  }

      if (mcxIOwarnfp(xf, "mcxIOopen PBD"))
      return STATUS_OK

   ;  if (strstr(xf->mode, "r"))
      treat   =  "reading"
   ;  else if (strstr(xf->mode, "w"))
      treat   =  "writing"
   ;  else if (strstr(xf->mode, "a"))
      treat   =  "appending"

   ;  if
      (  strstr(xf->mode, "r")
      && !strcmp(fname, "-")
      )
      {  xf->fp      =  stdin
      ;  xf->stdio   =  1
   ;  }
      else if
      (  (strstr(xf->mode, "w") || strstr(xf->mode, "a"))
      && !strcmp(fname, "-")
      )
      {  xf->fp      =  stdout
      ;  xf->stdio   =  1
   ;  }
      else if (!strcmp(fname, "stderr"))
      {  xf->fp      =  stderr
      ;  xf->stdio   =  1
   ;  }
      else if ((xf->fp = fopen(fname, xf->mode)) == NULL)
      {  if (ON_FAIL == RETURN_ON_FAIL)
         return STATUS_FAIL
      ;  else
            mcxIOerr(xf, "mcxIOopen", "can not be opened")
         ,  mcxExit(1)
   ;  }
      return STATUS_OK
;  }


void mcxIOrelease
(  mcxIO*  xf
)
   {  if (xf)
      {  mcxIOclose(xf)

      ;  if (xf->fn)
         mcxTingFree(&(xf->fn))
   ;  }
   }


void mcxIOappendName
(  mcxIO*       xf
,  const char*    suffix
)
   {  if (!xf || !suffix)
      {  mcxErr("mcxIOappendName PBD", "received void object")
      ;  mcxFail()
   ;  }
      else if (xf->fp)
      mcxErr
      (  "mcxIOappendName"
      ,  "stream open while request for name change from <%s> to <%s>"
      ,  xf->fn->str
      ,  suffix
      )
   ;  else if (!xf->fn)
      {  mcxErr("mcxIOappendName","no file name! for suffix <%s>",suffix)
      ;  mcxFail()
   ;  }
      else
      mcxTingAppend(xf->fn, suffix)
;  }


void mcxIOnewName
(  mcxIO*       xf
,  const char*    newname
)
   {  if (!xf || !newname)
      {  mcxErr("mcxIOnewName", "pbd: received void object")
      ;  mcxFail()
   ;  }
      else if (xf->fp)
      mcxErr
      (  "mcxIOnewName"
      ,  "stream open while request for name change from <%s> to <%s>"
      ,  xf->fn->str
      ,  newname
      )
   ;  else if (!xf->fn)
      {  mcxErr("mcxIOnewName", "no old file name! for new name <%s>", newname)
      ;  mcxFail()
   ;  }
      else
      mcxTingWrite(xf->fn, newname)
;  }


int mcxIOstepback
(  int c
,  mcxIO*    xf
)
   {  if (c == EOF)
      {  xf->ateof = 0
      ;  return EOF
   ;  }
      else
      {  if (ungetc(c, xf->fp) == EOF)
         {  mcxErr
            (  "mcxIOstepback"
            ,  "failed to push back <%d> on stream <%s>\n"
            ,  c
            ,  xf->fn->str
            )
         ;  return EOF
      ;  }

         xf->bc--
      ;  if (c == '\n')
         xf->lc--, xf->lo = 1  /* this is now semi-legally-corrupt */
      ;  else
         xf->lo--
   ;  }
      return c
;  }


int mcxIOstep
(  mcxIO*    xf
)
   {  int c = xf->ateof ? EOF : fgetc(xf->fp)
   ;  switch(c)
      {
      case '\n'
      :  xf->lc++
      ;  xf->bc++
      ;  xf->lo = 0
      ;  break
      ;

      case EOF
      :  xf->ateof =  1
      ;  break
      ;

      default
      :  xf->bc++
      ;  xf->lo++
      ;  break
   ;  }
      return c
;  }


mcxTing*  mcxIOreadFile
(  mcxIO    *xf
,  mcxTing   *filetxt
)
#define MCX_IORF_BSZ 513
   {  char cbuf[MCX_IORF_BSZ]
   ;  short c
   ;  int i =  0
   ;  struct stat mystat
   ;  size_t sz = 1024

   ;  if (!xf->stdio)
      {  if (stat(xf->fn->str, &mystat))
         mcxIOerr(xf, "mcxIOreadFile", "!can not stat file")
      ;  else
         sz = mystat.st_size
   ;  }

      if (!xf->fp && mcxIOopen(xf, RETURN_ON_FAIL))
      {  mcxIOerr(xf, "mcxIOreadFile", "can not open file")
      ;  return NULL
   ;  }

      if (xf->ateof)
      return mcxTingEmpty(filetxt, 0)   /* testme; 0 arg ok ? */

                                           /* fixme; ting count overflow */
   ;  filetxt = mcxTingEmpty(filetxt, sz)
   ;  if(0)
      fprintf(stderr, "allocing %d bytes\n", filetxt->mxl)

   ;  while
      (  c = mcxIOstep(xf)
      ,  c != EOF
      )
      {  cbuf[i]        =  (char) c

      ;  if (i+1 == MCX_IORF_BSZ)
         {  mcxTingNAppend(filetxt, cbuf, i+1)
         ;  i = 0
         ;  continue
      ;  }
         i++
   ;  }

      if (i)               /* nothing written at this position */
      mcxTingNAppend(filetxt, cbuf, i)

   ;  if (0) fprintf(stderr, "byte count %d\n", xf->bc)

   ;  xf->ateof   =  1

   ;  return filetxt
#undef MCX_IORF_BSZ
;  }


int  mcxIO__rl_fillbuf__
(  mcxIO*   xf
,  char*    buf
,  int      size
,  int*     last
)
   {  int   a  = 0
   ;  int   ct = 0
   ;  while(ct<size)          /* fixme; introduces extra newline at EOF! */
      {  buf[ct++] = ((a = mcxIOstep(xf)) == EOF) ? '\n' : a
      ;  if (a == '\n' || a == EOF)
         break
   ;  }
      *last =  a
   ;  return ct
;  }


int mcxIO__rl_rl__
(  mcxIO    *xf
,  mcxTing  *lineTxt
)
#define MCX_IORL_BSZ 513
   {  char  cbuf[MCX_IORL_BSZ]
   ;  int   z

   ;  mcxTingEmpty(lineTxt, 1)
   ;  while (1)
      {  int ct = mcxIO__rl_fillbuf__(xf, cbuf, MCX_IORL_BSZ, &z)
      ;  if (ct)
         mcxTingNAppend(lineTxt, cbuf, ct)
      ;  if (z == '\n' || z == EOF)
         break
   ;  }
      return z
#undef MCX_IORL_BSZ
;  }


int mcxIOdiscardLine
(  mcxIO       *xf
)
   {  int a, ct = 0

   ;  if (!xf->fp)
      {  mcxIOerr(xf, "mcxIOdiscardLine", "is not open")
      ;  return -1
   ;  }

      while(((a = mcxIOstep(xf)) != '\n') && a != EOF)
      ct++

   ;  return ct
;  }


mcxTing*  mcxIOreadLine
(  mcxIO          *xf
,  mcxTing        *dst
,  mcxbits        flags
)
   {  int      a, ll
   ;  mcxbool  chomp    =  flags & MCX_READLINE_CHOMP
   ;  mcxbool  skip     =  flags & MCX_READLINE_SKIP_EMPTY
   ;  mcxbool  par      =  flags & MCX_READLINE_PAR
   ;  mcxbool  dot      =  flags & MCX_READLINE_DOT
   ;  mcxbool  bsc      =  flags & MCX_READLINE_BSC
   ;  mcxbool  repeat   =  dot || par || bsc
   ;  mcxbool  continuation   =  FALSE
   ;  mcxTing* line

   ;  if (!xf->fp && mcxIOopen(xf, RETURN_ON_FAIL))
      {  mcxIOerr(xf, "mcxIOreadLine", "is not open")
      ;  return NULL
   ;  }

      if (xf->ateof)
      {  mcxTingFree(&dst)
      ;  return NULL
   ;  }

      dst = mcxTingEmpty(dst, 1)

   ;  if (skip || par)
      {  while((a = mcxIOstep(xf)) == '\n')
      /* NOTHING */
      ;  if (xf->ateof)
         {  mcxTingFree(&dst)
         ;  return NULL
      ;  }
         else
         mcxIOstepback(a, xf)
   ;  }

      line  =  repeat ? mcxTingEmpty(NULL, 1) : dst

   ;  while (1)
      {  a = mcxIO__rl_rl__(xf, line)
      ;  ll = line->len

      ;  if (!repeat)
         break
      ;  else  /* must append line to dst */
         {  if (dot && !continuation && ll == 2 && *(line->str) == '.')
            break
         /* do not attach the single-dot-line */

         ;  if (par && !continuation && ll == 1)
            break
         /* do not attach the second newline */

         ;  mcxTingNAppend(dst, line->str, line->len)

         ;  continuation = bsc && (ll > 1 && *(line->str+ll-2) == '\\')

         ;  if (continuation)
            mcxTingShrink(dst, -2)

         ;  if (!par && !dot && (bsc && !continuation))
            break

         ;  if (xf->ateof)
            break
      ;  }
      }

      if (chomp && *(dst->str+dst->len-1) == '\n')
      mcxTingShrink(dst, -1)

   ;  if (repeat)
      mcxTingFree(&line)

   ;  if (xf->ateof && !dst->len)
      mcxTingFree(&dst)

   ;  return dst
;  }


void mcxIOlistParmodes
(  void
)
   {  fprintf
      (  stdout
      ,  "%5d  vanilla mode, fetch next line\n"
      ,  MCX_READLINE_DEFAULT
      )
   ;  fprintf(stdout, "%5d  chomp trailing newline\n", MCX_READLINE_CHOMP)
   ;  fprintf(stdout, "%5d  skip empty lines\n", MCX_READLINE_SKIP_EMPTY)
   ;  fprintf(stdout, "%5d  paragraph mode\n", MCX_READLINE_PAR)
   ;  fprintf(stdout, "%5d  backslash escapes newline\n", MCX_READLINE_BSC)
   ;  fprintf
      (  stdout
      ,  "%5d  section mode, ended by singly dot on a single line\n"
      ,  MCX_READLINE_DOT
      )
;  }


void mcxIOpos
(  mcxIO*    xf
,  FILE*          channel
)
   {  const char* ateof =  xf->ateof ? "at EOF in " : ""
   ;  fprintf
      (  channel
      ,  "[mclIO] %sstream <%s>, line <%d>, character <%d>\n"
      ,  ateof
      ,  xf->fn->str
      ,  xf->lc
      ,  xf->lo
      )
;  }


void  mcxIOrewind
(  mcxIO*   xf
)
   {  xf->lc   =  1
   ;  xf->lo   =  0
   ;  xf->bc   =  0
;  }


void mcxIOfree_v
(  void*  xfpp
)  
   {  mcxIOfree((mcxIO**) xfpp)
;  }


void mcxIOfree
(  mcxIO**  xf
)  
   {  if (*xf)
      {  mcxIOrelease(*xf)
      ;  mcxFree(*xf)
      ;  *xf   =  NULL
   ;  }
   }


/* fixme: newlines in str thrash correct lc counting */
/* fixme: offset not reset. consider e.g. search for strstrx in strstrstrx */

mcxstatus mcxIOexpect
(  mcxIO                *xf
,  const char           *str
,  mcxOnFail            ON_FAIL
)
   {  const char*       s        =  str
   ;  short             c        =  0
   ;  short             d        =  0
   ;  long              bct      =  0
   ;  FILE*             fp       =  xf->fp
   ;  long              pos      =  ftell(fp)
   ;  int               reset    =  !xf->stdio

   ;  while
      (  c = s[0]
      ,  
         (  c
         && (  d = fgetc(fp)
            ,  bct++
            ,  c == d
            )  
         )  
      )
      s++

   ;  if (c)
      {  if (ON_FAIL == EXIT_ON_FAIL)
         {  mcxErr("mcx parse", "parse error: expected to see <%s>", str)
         ;  mcxIOpos(xf, stderr)
         ;  mcxExit(1)
      ;  }
         else
         {  if (reset)
            fseek(fp, pos, SEEK_SET)
         ;  else
            {  ungetc(d, fp)     /* put back first nonmatching character */
            ;  xf->bc += bct-1   /* note that bct >= 1 since c != '\0'    */
            ;  xf->lo += bct-1   /* fixbug: no newlines in pattern ??? */
         ;  }
         }
         return STATUS_FAIL
   ;  }
      else
      {  xf->bc += bct
      ;  xf->lo += bct
      ;  return STATUS_OK
   ;  }
;  }


mcxstatus mcxIOexpectReal
(  mcxIO*        xf
,  double*       dblp
,  mcxOnFail     ON_FAIL
)
   {  int   n_read   =  0
   ;  int   n_conv   =  fscanf(xf->fp, " %lf%n", dblp, &n_read)

   ;  xf->bc += n_read  /* mq meaninful in case of error ? */
   ;  xf->lo += n_read

   ;  if (1 != n_conv)
      {  mcxIOpos(xf, stderr)
      ;  mcxErr("parseReal", "parse error: expected to find real")
      ;  if (ON_FAIL == EXIT_ON_FAIL)
         mcxExit(1)
      ;  return STATUS_FAIL
   ;  }
      return STATUS_OK
;  }


mcxstatus mcxIOexpectNum
(  mcxIO*        xf
,  long*         lngp
,  mcxOnFail     ON_FAIL
)
   {  int   n_read   =  0
   ;  int   n_conv   =  fscanf(xf->fp, "%ld%n", lngp, &n_read)

   ;  xf->bc += n_read  /* fixme meaninful in case of error ? */
   ;  xf->lo += n_read

   ;  if (1 != n_conv)
      {  mcxIOpos(xf, stderr)
      ;  mcxErr("parseNum", "parse error: expected to find integer")
      ;  if (ON_FAIL == EXIT_ON_FAIL)
         mcxExit(1)
      ;  return STATUS_FAIL
   ;  }
      return STATUS_OK
;  }



int mcxIOskipSpace
(  mcxIO*               xf
)
   {  short             c
   ;  long              lct    =    0
   ;  long              bct    =    0
   ;  long              lo     =    xf->lo

   ;  while ((c = fgetc(xf->fp)) != EOF)
      {  if (c=='\n')
         {  lct++
         ;  bct++
         ;  lo   =  1
      ;  }
         else if (isspace(c))
         {  bct++
         ;  lo++
      ;  }
         else
         {  break
      ;  }
   ;  }

   ;  xf->lc +=  lct
   ;  xf->bc +=  bct
   ;  xf->lo  =  lo

   ;  if (c == EOF)
      {  xf->ateof   =  1
      ;  return EOF
   ;  }
      else
      {  ungetc(c, xf->fp)
      ;  return c
   ;  }
;  }


mcxstatus mcxIOfind
(  mcxIO*         xf
,  const char*    str
,  mcxOnFail      ON_FAIL
)
   {  short c

   ;  while(1)
      {  c = mcxIOstep(xf)

      ;  if (c == EOF)
         {  mcxErr
            (  "mcx find-in-file"
            ,  "in file <%s> at EOF, expected <%s> not found"
            ,  xf->fn->str
            ,  str
            )
         ;  mcxIOpos(xf, stderr)
         ;  if (ON_FAIL == RETURN_ON_FAIL)
            return STATUS_FAIL
         ;  else
            mcxExit(1)
      ;  }

         if
         (  c == str[0] 
         && (  mcxIOexpect
               (  xf
               ,  str + 1
               ,  RETURN_ON_FAIL
               )
            == STATUS_OK
            )
         )
         return STATUS_OK
   ;  }
      return STATUS_FAIL
;  }



