/*            Copyright (C) 1999, 2000, 2001, 2002, 2003 Stijn van Dongen
 *
 * This file is part of MCL.  You can redistribute and/or modify MCL under the
 * terms of the GNU General Public License; either version 2 of the License or
 * (at your option) any later version.  You should have received a copy of the
 * GPL along with MCL, in the file COPYING.
*/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <math.h>

#include "compile.h"
#include "ting.h"
#include "ding.h"
#include "minmax.h"
#include "alloc.h"
#include "array.h"
#include "hash.h"
#include "compile.h"
#include "types.h"
#include "err.h"
#include "pool.h"


#if !MCX_UTIL_THREADED
static MP* tingPool_g = NULL;
#endif


void mcxTingAlert        
(  const char*          caller
)
   {  mcxErr(caller, "void ting argument")
   ;  return
;  }


void mcxTingRelease
(  void* tingv
)
   {  if (tingv && ((mcxTing*)tingv)->str)
      mcxFree(((mcxTing*)tingv)->str)
;  }


void mcxTingFree_v
(  void  *tingpp
)
   {  mcxTingFree((mcxTing**) tingpp)
;  }


void mcxTingFree
(  mcxTing            **tingpp
)
   {  mcxTing*  ting   =    *tingpp

   ;  if (ting)
      {  
         if (ting->str)
         mcxFree(ting->str)

#if MCX_UTIL_THREADED
      ;  mcxFree(ting)
#else
      ;  mp_free(tingPool_g, ting)
#endif
      ;  *tingpp = NULL
   ;  }
;  }


void mcxTingShrink
(  mcxTing*    t
,  int         len
)
   {  if (!t)
         mcxTingAlert("mcxTingShrink")
      ,  mcxFail()

   ;  if (len < 0)
      len = t->len + len

   ;  if (len < 0 || len > t->len)
      {  mcxErr("mcxTingShrink","funny len <%d> newlen <%d> combo",t->len,len)
      ;  return
   ;  }
      else
      {  *(t->str+len) = '\0'
      ;  t->len = len;
   ;  }
   }


mcxTing* mcxTingEnsure
(  mcxTing*    ting
,  int         len
)  
   {  if (!ting)
      ting  =  (mcxTing*) mcxTingInit(NULL)
   
   ;  if (len < 0)
         mcxErr("mcxTingEnsure", "pbd: request for negative length <%d>", len)
      ,  mcxFail()
   ;  else if (len <= ting->mxl)
   ;  else
      {  ting->str = (char*) mcxRealloc(ting->str, len+1, RETURN_ON_FAIL)

      ;  if (!ting->str)
            mcxMemDenied(stderr, "mcxTingEnsure", "char", len)
         ,  mcxFail()

      ;  ting->mxl =  len
      ;  if (1)
         *(ting->str+ting->mxl) =  '\0'
         /* should be no need to do this;
          * and for large allocations this may fail.
         */
   ;  }

      return ting
;  }


mcxTing*  mcxTing_print_
(  mcxTing*    dst
,  const char* fmt
,  va_list     *args
)
#define PRINT_BUF_SIZE 200
   {  
      char buf[PRINT_BUF_SIZE]
   ;  mcxTing* txtbuf = NULL  
   ;  char* src
   ;  int m = PRINT_BUF_SIZE
   ;  int n

   ;  n = vsnprintf(buf, PRINT_BUF_SIZE, fmt, *args)

   ;  if (n<0 || n >= PRINT_BUF_SIZE)
      {  
         m  =  n >= PRINT_BUF_SIZE ? n+1 : 2*m
      ;  while (1)
         {  txtbuf = mcxTingEmpty(txtbuf, m)
         ;  n = vsnprintf(txtbuf->str, m+1, fmt, *args)
         ;  if (n<0 || n > m)
            m *= 2
         ;  else
            break
         ;  if (m > 1040000)
               mcxErr("mcxTingPrint", "exceeding compile time 1040K limit")
            ,  mcxExit(1)
      ;  }
         src = txtbuf->str
   ;  }
      else
      src = buf

   ;  if (!dst)
      dst = mcxTingNew(src)
   ;  else
      mcxTingWrite(dst, src)

   ;  mcxTingFree(&txtbuf)
   ;  return dst
;  }


mcxTing*  mcxTingPrint
(  mcxTing*    dst
,  const char* fmt
,  ...
)
   {  va_list  args

   ;  va_start(args, fmt)
   ;  dst = mcxTing_print_(dst, fmt, &args)
   ;  va_end(args)
   ;  return dst
;  }


mcxTing*  mcxTingPrintAfter
(  mcxTing*    dst
,  const char* fmt
,  ...
)
   {  va_list  args
   ;  mcxTing *affix = NULL

   ;  va_start(args, fmt)
   ;  affix = mcxTing_print_(affix, fmt, &args)
   ;  va_end(args)

   ;  if (!dst)
      return affix

   ;  mcxTingAppend(dst, affix->str)
   ;  mcxTingFree(&affix)
   ;  return dst
;  }


static char* roman[40] =
{  "" , "i" , "ii", "iii", "iv"  , "v" , "vi", "vii", "viii", "ix"
,  "" , "x" , "xx", "xxx", "xl"  , "l" , "lx", "lxx", "lxxx", "xc"
,  "" , "c" , "cc", "ccc", "cd"  , "d" , "dc", "dcc", "dccc", "cm"
,  "" , "m" , "mm", "mmm", "mmmm", ""  ,  "" ,    "",     "", ""
}  ;


mcxTing*  mcxTingRoman
(  mcxTing*  dst
,  int       a
,  int       ucase
)
   {  int i, x, c, m
   ;  char* p

   ;  if (a >= 5000 || a <= 0)
      return (dst ? mcxTingWrite(dst, "-"), dst : mcxTingNew("-"))

   ;  i  =  a % 10
   ;  a /=  10
   ;  x  =  a % 10
   ;  a /=  10
   ;  c  =  a % 10
   ;  a /=  10
   ;  m  =  a

   ;  dst = mcxTingPrint
      (  dst
      ,  "%s%s%s%s"
      ,  roman[30+m]
      ,  roman[20+c]
      ,  roman[10+x]
      ,  roman[ 0+i]
      )
   ;  if (ucase)
      for (p=dst->str;p<dst->str+dst->len;p++)
      *p += 'A' - 'a'

   ;  return dst
;  }


mcxTing*  mcxTingDouble
(  mcxTing* dst
,  double   x
,  int      decimals
)
   {  char num[500]
   ;  char* p
   ;  int len

   ;  if (decimals < 0)
         mcxErr("mcxTingDouble PBD", "negative decimals arg")
      ,  mcxExit(1)

  /* mq should check return value and use snprintf */
   ;  len =  sprintf(num, "%.*f", decimals, x) -1
   ;  p   =  num+len-1
                                 /* GNU libcx bug workaround         */
                                 /* it returns 2 for length of 'inf' */ 
   ;  if (decimals && strcmp(num, "inf"))
      {  while (*p == '0')
         p--
      ;  if (*p == '.')
         *++p = '0'
      ;  *++p = '\0'
   ;  }
      if (dst)
      {  mcxTingWrite(dst, num)
      ;  return dst
   ;  }

      return mcxTingNew(num)
;  }


mcxTing*  mcxTingInteger
(  mcxTing*  dst
,  long     x
)
   {  char num[28]      /* perfect! */

   ;  sprintf(num, "%ld", x)
   ;  if (dst)
      {  mcxTingWrite(dst, num)
      ;  return dst
   ;  }

   ;  return mcxTingNew(num)
;  }


void* mcxTingInit
(  void *  tingv
)
   {  mcxTing *ting  =  (mcxTing*) tingv

   ;  if (!ting)
      {
#if MCX_UTIL_THREADED
         ting =  (mcxTing*) mcxAlloc(sizeof(mcxTing), EXIT_ON_FAIL)
#else
         if (!tingPool_g)
         tingPool_g = mp_init(4096, MP_EXPONENTIAL, sizeof(mcxTing), 0)
      ;  ting = (mcxTing*) mp_alloc(tingPool_g)
#endif
   ;  }

   ;  ting->str      =  (char*) mcxAlloc(sizeof(char), EXIT_ON_FAIL)
   ;  *(ting->str+0) =  '\0'
   ;  ting->len      =  0
   ;  ting->mxl      =  0

   ;  return  ting
;  }


/*
 *    Take string into an existing ting or into a new ting
*/

mcxTing* mcxTingInstantiate
(  mcxTing*          ting
,  const char*       string
)
   {  int length =   string ? strlen(string) : 0

   ;  ting = mcxTingEnsure(ting, length)     /* handles ting==NULL case */

	;  if (string)
      {  strncpy(ting->str, string, length)
      ;  *(ting->str+length)  =  '\0'
   ;  }

      ting->len = length
	;  return ting
;  }


int mcxTingCmp
(  const void* t1
,  const void* t2
)
   {  return (strcmp(((mcxTing*)t1)->str, ((mcxTing*)t2)->str))
;  }


void mcxTingSplice
(  mcxTing*       ting
,  const char*    pstr
,  int            d_offset
,  int            n_delete
,  int            n_copy
)
   {  int      newlen   =  ting->len - n_delete + n_copy

   ;  if (ting == NULL)
         mcxTingAlert("mcxTingSplice")
      ,  mcxFail()

   ;  if (newlen < 0)
      {  mcxErr("mcxTingSplice", "pbd: arguments result in negative length")
      ;  return 
   ;  }

  /*
   *  essential: mcxSplice does not know to allocate room for '\0'
  */
   ;  mcxTingEnsure(ting, newlen)

   ;  mcxSplice
      (  &(ting->str)
      ,  pstr
      ,  sizeof(char)
      ,  &(ting->len)
      ,  &(ting->mxl)
      ,  d_offset
      ,  n_delete
      ,  n_copy
      )

  /*
   *  essential: mcxSplice might have realloced, so has to be done afterwards.
  */
   ;  *(ting->str+newlen)  =  '\0'

   ;  if (ting->len != newlen)
         mcxErr("mcxSpliceting", "pbd: mcxSplice gives unexpected result")
      ,  mcxFail()
;  }


mcxTing* mcxTingNew
(  const char* str
)
   {  return mcxTingInstantiate(NULL, str)
;  }


mcxTing* mcxTingNNew
(  const char*       str
,  int               n
)
   {  mcxTing*  ting   =  mcxTingEnsure(NULL, n)
   ;  strncpy(ting->str, str, n)
   ;  *(ting->str+n) = '\0'
   ;  ting->len = n
   ;  return ting
;  }


mcxTing* mcxTingEmpty
(  mcxTing*           ting
,  int               len
)  {
      ting               =   mcxTingEnsure(ting, len)
   ;  *(ting->str+0)     =  '\0'
   ,  ting->len          =  0
   ;  return ting
;  }


mcxTing* mcxTingWrite
(  mcxTing*          ting
,  const char*       str
)
   {  return mcxTingInstantiate(ting, str)
;  }


mcxTing* mcxTingNWrite
(  mcxTing*          ting
,  const char*       str
,  int               n
)
   {  ting = mcxTingEnsure(ting, n)
   ;  strncpy(ting->str, str, n)
   ;  *(ting->str+n) = '\0'
   ;  ting->len = n
   ;  return ting
;  }


char* mcxTingSubStr
(  mcxTing*    ting
,  int         offset  
,  int         length
)
   {  char* str
   ;  if (offset < 0 || offset > ting->len)
      offset = ting->len
   ;  if (length < 0 || offset+length > ting->len)
      length = ting->len - offset
      
   ;  str = mcxAlloc(length+1, EXIT_ON_FAIL)
   ;  if (length)
      memcpy(str, ting->str+offset, length)         
   ;  *(str+length) = '\0'
   ;  return str
;  }


char* mcxTingStr
(  mcxTing*           ting
)
   {  return mcxTingSubStr(ting, 0, ting->len)
;  }



mcxTing* mcxTingAppend
(  mcxTing*              ting
,  const char*          str
)
   {  if (!ting)
      return mcxTingNew(str)

   ;  mcxTingSplice
      (  ting
      ,  str
      , -1                             /*    splice offset     */
      ,  0                             /*    delete nothing    */
      ,  str? strlen(str) : 0          /*    string length     */
      )
   ;  return ting
;  }


mcxTing* mcxTingKAppend
(  mcxTing*    ting
,  const char* str
,  int         n
)
   {  int len = strlen(str)
   ;  while (n--)
      ting = mcxTingNAppend(ting, str, len)

   ;  return ting
;  }


mcxTing* mcxTingNAppend
(  mcxTing*              ting
,  const char*          str
,  int                  n
)
   {  if (!ting)
      return mcxTingNWrite(NULL, str, n)

   ;  mcxTingSplice
      (  ting
      ,  str
      , -1                             /*    splice offset     */
      ,  0                             /*    delete nothing    */
      ,  n                             /*    string length     */
      )
   ;  return ting
;  }


void mcxTingInsert
(  mcxTing*              ting
,  const char*          str
,  int                  offset
)
   {  mcxTingSplice
      (  ting
      ,  str
      ,  offset                        /*    splice offset     */
      ,  0                             /*    delete nothing    */
      ,  str? strlen(str) : 0          /*    string length     */
      )
;  }


void mcxTingNInsert
(  mcxTing*          ting
,  const char*       str
,  int               offset
,  int               length
)
   {  mcxTingSplice
      (  ting
      ,  str
      ,  offset                        /*    splice offset     */
      ,  0                             /*    delete nothing    */
      ,  length                        /*    length of str     */
      )
;  }


void mcxTingDelete
(  mcxTing*          ting
,  int               offset
,  int               width
)
   {  mcxTingSplice
      (  ting
      ,  NULL
      ,  offset                        /*    splice offset     */
      ,  width                         /*    delete width      */
      ,  0                             /*    string length     */
      )
;  }


int mcxTingTranslate
(  mcxTing* src
,  int*     tbl
,  int      flags
)
   {  src->len  =  mcxStrTranslate(src->str, tbl, flags)
   ;  return src->len
;  }


int mcxTingTr
(  mcxTing*    txt
,  const char* src
,  const char* dst
,  int         flags
)
   {  int tbl[256]
   ;  if (!trLoadTable(src, dst, tbl, flags))
      return -1
   ;  txt->len  =  mcxStrTranslate(txt->str, tbl, flags)
   ;  return txt->len
;  }



/* ================================================================== *
 *
 * ================================================================== *
*/


u32 mcxTingDJBhash
(  const void* ting
)
   {  return(mcxBJhash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingBDBhash
(  const void* ting
)
   {  return(mcxBDBhash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingBJhash
(  const void* ting
)
   {  return(mcxBJhash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingCThash
(  const void* ting
)
   {  return(mcxCThash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingDPhash
(  const void* ting
)
   {  return(mcxDPhash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingGEhash
(  const void* ting
)
   {  return(mcxGEhash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingOAThash
(  const void* ting
)
   {  return(mcxOAThash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingELFhash
(  const void* ting
)
   {  return(mcxELFhash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 (*mcxTingHFieByName(const char* id))(const void* ting)
   {  if (!strcmp(id, "dp"))
      return mcxTingDPhash
   ;  else if (!strcmp(id, "bj"))
      return mcxTingBJhash
   ;  else if (!strcmp(id, "elf"))
      return mcxTingELFhash
   ;  else if (!strcmp(id, "djb"))
      return mcxTingDJBhash
   ;  else if (!strcmp(id, "bdb"))
      return mcxTingBDBhash
   ;  else if (!strcmp(id, "ge"))
      return mcxTingGEhash
   ;  else if (!strcmp(id, "oat"))
      return mcxTingOAThash
   ;  else if (!strcmp(id, "svd"))
      return mcxTingSvDhash
   ;  else if (!strcmp(id, "svd2"))
      return mcxTingSvD2hash
   ;  else if (!strcmp(id, "svd1"))
      return mcxTingSvD1hash
   ;  else if (!strcmp(id, "ct"))
      return mcxTingCThash
   ;  else
      return NULL
;  }


u32 mcxTingSvDhash
(  const void* ting
)
   {  return(mcxSvDhash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingSvD2hash
(  const void* ting
)
   {  return(mcxSvD2hash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingSvD1hash
(  const void* ting
)
   {  return(mcxSvD1hash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }


u32 mcxTingHash
(  const void* ting
)
   {  return(mcxDPhash(((mcxTing*) ting)->str, ((mcxTing*) ting)->len))
;  }

