/*            Copyright (C) 1999, 2000, 2001, 2002, 2003 Stijn van Dongen
 *
 * This file is part of MCL.  You can redistribute and/or modify MCL under the
 * terms of the GNU General Public License; either version 2 of the License or
 * (at your option) any later version.  You should have received a copy of the
 * GPL along with MCL, in the file COPYING.
*/

#ifndef util_ting
#define util_ting

/* ========================================================================= *
 *
 *    str:  array of chars
 *
 *    len:  length of first string in str (excluding '\0')
 *          *(str+len) == '\0'
 *
 *    mxl:  current allocated number of writable char's (excluding '\0')
 *          Allocated amount is mxl+1
 *          *(str+mxl) == '\0' (umm, is that really true?)
 *
 *    mcxTingEnsure is the only routine allowed to fiddle with mxl.
 *    (apart from mcxTingInit which sets it to zero).
 *    Future option could be something like mcxTingFinalize,
 *    releasing unused memory.
 * 
 *    Routines marked `+' and `#' will accept ting==NULL argument.
 *    Routines marked `#' should not be called other than by ting routines
 * 
 *                  #tingInit
 *                  /       \
 *          #tingEnsure <- #tingInstantiate
 *          :      \               :                   
 *  +Empty        tingSplice   +tingNew,Write
 *   NWrite           :               
 *                    :
 *           tingAppend,Insert,Delete
 *
 * ========================================================================= *
*/


#include <string.h>
#include "alloc.h"
#include "types.h"
#include "hash.h"


typedef struct
{  char     *str
;  int      len
;  int      mxl
;
}  mcxTing  ;


/* ========================================================================= *
 *
 *    various instantiation routines.
 *
 * ========================================================================= *
*/

                          /*     Accepts NULL argument.
                          */
mcxTing* mcxTingEnsure
(  mcxTing*    ting
,  int         length
)  ;  


                          /*     Accepts NULL argument.
                          */
mcxTing* mcxTingEmpty
(  mcxTing*    ting
,  int         len
)  ;


                          /*     Accepts NULL argument.
                          */
mcxTing* mcxTingPrint
(  mcxTing*    ting
,  const char* fmt
,  ...
)
#ifdef __GNUC__
__attribute__ ((format (printf, 2, 3)))
#endif
   ;

                          /*     Accepts NULL argument.
                          */
mcxTing*  mcxTingPrintAfter
(  mcxTing*    dst
,  const char* fmt
,  ...
)
#ifdef __GNUC__
__attribute__ ((format (printf, 2, 3)))
#endif
   ;

                          /*     accepts NULL argument, maps to empty string.
                          */
mcxTing* mcxTingNew
(  const char* str
)  ;

                          /*     accepts NULL argument, maps to empty string.
                          */
mcxTing* mcxTingNNew
(  const char* str
,  int         n
)  ;

                          /*     Accepts NULL argument.
                           *     void arguments so that it can be passed.
                          */
void* mcxTingInit
(  void* ting
)  ;  

                          /*     Accepts NULL argument.
                          */
mcxTing* mcxTingInstantiate
(  mcxTing*    dst_ting
,  const char* str
)  ;  


/* ========================================================================= *
 *
 *    freeing.
 *
 * ========================================================================= *
*/


void mcxTingFree
(  mcxTing     **tingpp
)  ;

                          /*     use with generic callbacks, e.g.
                           *     for freeing hash with tings in one go.
                          */
void mcxTingFree_v
(  void        *tingpp
)  ;


void mcxTingRelease
(  void        *ting
)  ;


/* ========================================================================= *
 *
 *    writing, shrinking, inserting, appending, deleting.
 *
 * ========================================================================= *
*/

                          /*     does NOT accept NULL argument.
                          */
void mcxTingSplice
(  mcxTing*    ting
,  const char* pstr
,  int         d_offset
,  int         n_delete
,  int         n_copy
)  ;

                          /*     does NOT accept NULL argument.
                          */
void mcxTingInsert
(  mcxTing*    ting
,  const char* str
,  int         offset
)  ;

                          /*     does NOT accept NULL argument.
                          */
void mcxTingNInsert
(  mcxTing*    ting
,  const char* str
,  int         offset     /* of ting->str */
,  int         length     /* of str       */
)  ;

                          /*     accepts NULL argument.
                          */
mcxTing* mcxTingAppend
(  mcxTing*    ting
,  const char* str
)  ;

                          /*     accepts NULL argument.
                          */
mcxTing* mcxTingKAppend
(  mcxTing*    ting
,  const char* str
,  int         n
)  ;

                          /*     accepts NULL argument.
                          */
mcxTing* mcxTingNAppend
(  mcxTing*    ting
,  const char* str
,  int         n
)  ;

                          /*     does NOT accept NULL argument.
                          */
void mcxTingDelete
(  mcxTing*    ting
,  int         offset
,  int         width
)  ;

                          /*     accepts NULL argument.
                          */
mcxTing* mcxTingWrite
(  mcxTing*    ting
,  const char* str
)  ;

                          /*     accepts NULL argument.
                          */
mcxTing* mcxTingNWrite
(  mcxTing*    ting
,  const char* str
,  int         n
)  ;


void mcxTingShrink
(  mcxTing*    ting
,  int         length
)  ;


/* ========================================================================= *
 *
 *    miscellaneous/general.
 *
 * ========================================================================= *
*/


void mcxTingAlert        
(  const char*  caller
)  ;


char*  mcxTingStr
(  mcxTing*    ting
)  ;

char* mcxTingSubStr
(  mcxTing*    ting
,  int         offset
,  int         length
)  ;

int mcxTingTranslate
(  mcxTing*    src
,  int*        tbl        /* should have size 256 */
,  int         flags
)  ;

int mcxTingTr
(  mcxTing*    txt
,  const char* src
,  const char* dst
,  int         flags
)  ;

                          /*     accepts NULL argument.
                          */
mcxTing*  mcxTingRoman
(  mcxTing*    dst
,  int         x
,  int         ucase
)  ;

                          /*     accepts NULL argument.
                          */
mcxTing*  mcxTingInteger
(  mcxTing*    dst
,  long        x
)  ;


mcxTing*  mcxTingDouble
(  mcxTing* dst
,  double   x
,  int      decimals
)  ;


int mcxTingCmp
(  const void* t1
,  const void* t2
)  ;


/* ========================================================================= *
 *
 *    hashing.
 *
 * ========================================================================= *
*/

u32 (*mcxTingHFieByName(const char* id))(const void* ting)  ;


u32 mcxTingELFhash
(  const void* ting
)  ;

u32 mcxTingHash
(  const void* ting
)  ;

u32 mcxTingBJhash
(  const void* ting
)  ;

u32 mcxTingCThash
(  const void* ting
)  ;

u32 mcxTingDPhash
(  const void* ting
)  ;

u32 mcxTingBDBhash
(  const void* ting
)  ;

u32 mcxTingGEhash
(  const void* ting
)  ;

u32 mcxTingOAThash
(  const void* ting
)  ;

u32 mcxTingSvDhash
(  const void* ting
)  ;

u32 mcxTingSvD2hash
(  const void* ting
)  ;

u32 mcxTingSvD1hash
(  const void* ting
)  ;

u32 mcxTingDJBhash
(  const void* ting
)  ;

#endif

